#include "clause2.h"

/* Private definitions and types */

/*************
 *
 *   delete_clause()
 *
 *************/

/* DOCUMENTATION
This routine frees a clause and all of its subterms.
If the clause has an ID, it is unassigned.
If the clause has a justification list, it is freed.
<P>
This routine is not in the clause package, because
(at this time) the clause package doesn't know about
just.h or clauseid.h.
*/

/* PUBLIC */
void delete_clause(Clause c)
{
  zap_just(c->justification);  /* ok if NULL */
  if (has_clause_id(c))
    unassign_clause_id(c);
  zap_clause(c);
}  /* delete_clause */

/*************
 *
 *   delete_clist()
 *
 *************/

/* DOCUMENTATION
For each Clause in the Clist, remove it from the Clist,
and call delete_clause().  Finally, free the Clist.
<P>
This routine is not in the clist package, because
(at this time) the clist package doesn't know about
just.h or clauseid.h.
*/

/* PUBLIC */
void delete_clist(Clist l)
{
  Clist_pos p;
  Clause c;

  p = l->first;
  while (p) {
    c = p->c;
    p = p->next;
    clist_remove(c, l);
    if (c->containers == NULL)
      delete_clause(c);
  }
  clist_free(l);
}  /* delete_clist */

/*************
 *
 *   demod_clause()
 *
 *************/

/* DOCUMENTATION
Demodulate Clause c, using demodulators in Mindex idx.
If any rewriting occurs, the justification is appended to
the clause's existing justification.
*/

/* PUBLIC */
void demod_clause(Clause c, Mindex idx)
{
  if (mindex_empty(idx))
    return;
  else {
    Literal lit;
    Glist g = NULL;
    for (lit = c->first_literal; lit != NULL; lit = lit->next) {
      lit->atom = demodulate(lit->atom, idx, &g);
    }
    if (g != NULL) {
      upward_clause_links(c);
      g = reverse_glist(g);
      g = demod_just(g);
      c->justification = glist_cat(c->justification, g);
    }
  }
}  /* demod_clause */

