#include "detach.h"

/* Private definitions and types */

/*************
 *
 *   detach()
 *
 *************/

static
Clause detach(Clause c1, Clause c2)
{
  /* Assume each is positive unit in unary relation with binary arg.
   * Condensed detach with major premise c1 and minor premise c2.
   */
  Term major = c1->first_literal->atom->args[0];
  Term alpha = major->args[0];
  Term beta  = major->args[1];
  Term minor = c2->first_literal->atom->args[0];
  Context s1 = get_context();
  Context s2 = get_context();
  Trail tr = NULL;
  Clause result;

  if (unify(alpha, s1, minor, s2, &tr)) {
    Term atom = get_rigid_term_like(c2->first_literal->atom);
    Literal lit = new_literal(1, atom);

    atom->args[0] = apply(beta, s1);
    undo_subst(tr);
    result = get_clause();
    append_literal(result, lit);
    result->justification = cd_just(c1, c2);
    upward_clause_links(result);
  }
  else
    result = NULL;
  free_context(s1);
  free_context(s2);
  return result;
}  /* detach */

/*************
 *
 *   cond_detach()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Glist cond_detach(Clause c1, Clause c2)
{
  /* First make sure that both are positive units of the form p(i(a,b)),
   * for any p, i, a, b.
   */

  if (!unit_clause(c1) || !unit_clause(c2))
    return NULL;
  else if (!positive_clause(c1) || !positive_clause(c2))
    return NULL;
  else if (ARITY(c1->first_literal->atom) != 1 ||
	   ARITY(c2->first_literal->atom) != 1)
    return NULL;
  else if (ARITY(c1->first_literal->atom->args[0]) != 2 ||
	   ARITY(c2->first_literal->atom->args[0]) != 2)
    return NULL;
  else {
    Clause cd1;
    Clause cd2;
    Glist results = NULL;

    cd1 = detach(c1, c2);
    cd2 = (c1 == c2 ? NULL : detach(c2, c1));

    if (cd1 != NULL)
      results = plist_append(results, cd1);
    if (cd2 != NULL)
      results = plist_append(results, cd2);
    
    return results;
  }
}  /* cond_detach */
