#ifndef TP_GLIST_H
#define TP_GLIST_H

#include "memory.h"

/* INTRODUCTION
This package handles Glist (generic) lists, which are simple,
singly linked lists.
The members of the lists can be integers or pointers.
In particular, each is a union of (int) and (void *).
There is no way to determine the types of the members of the
list, so you have to remember what kind of data are there.
<P>
Note that when assigning to and from a (void *), you don't need
a cast, but when dereferencing a (void *), you do need a cast.
<P>
There are two informal subtypes: "ilist" for lists of integers,
and "plist" for lists of pointers.  Several of the routines
(which have "ilist" or "plist" in the name) apply to these
suptypes.  The rest of the routines apply to any Glist.
*/

/* Public definitions */

typedef struct glist * Glist;
typedef struct g2list * G2list;

struct glist {
  Glist next;
  union {
    int i;
    void *v;
  } u;
};

struct g2list {
  G2list next;
  union {
    int i;
    void *v;
  } u1;
  union {
    int i;
    void *v;
  } u2;
};

/* End of public definitions */

/* Public function prototypes from glist.c */

Glist get_glist(void);

void free_glist(Glist p);

G2list get_g2list(void);

void free_g2list(G2list p);

void fprint_glist_mem(FILE *fp, BOOL heading);

void p_glist_mem(void);

Glist reverse_glist(Glist p);

Glist copy_glist(Glist p);

void zap_glist(Glist p);

int glist_count(Glist p);

void fprint_ilist(FILE *fp, Glist p);

void p_ilist(Glist p);

Glist ilist_append(Glist lst, int i);

Glist ilist_prepend(Glist lst, int i);

BOOL ilist_member(Glist lst, int i);

Glist plist_append(Glist lst, void *p);

Glist plist_prepend(Glist lst, void *p);

BOOL plist_member(Glist lst, void *p);

Glist glist_pop(Glist p);

Glist glist_cat(Glist p1, Glist p2);

void zap_g2list(G2list p);

G2list i2list_prepend(G2list lst, int i1, int i2);

G2list g2list_pop(G2list p);

#endif  /* conditional compilation of whole file */
