#ifndef TP_MACE_H
#define TP_MACE_H

#include "parse.h"
#include "formula.h"
#include "putnam.h"
#include "ioutil.h"

/* INTRODUCTION
This package looks for (small) finite models of (unsorted first order)
clause sets.  You call mace_first_model() with a list of clauses
and a bunch of parameters, and if it succeeds, it returns a
term representation of the model.  Also see the package <I>interp</I>,
which allows you to do things with the model, such as print it
and evaluate clauses and formulas w.r.t. it.
*/

/* Public definitions */

#define MAX_MACE_RELATION_ARITY   4
#define MAX_MACE_FUNCTION_ARITY   3

typedef struct mace_state * Mace_state;

/* End of public definitions */

/* Public function prototypes from mace.c */

void fprint_mace_mem(FILE *fp, BOOL heading);

void p_mace_mem();

void print_model(FILE *fp);

Term mace_first_model(Clist lst,
		      int domain_size,
		      BOOL distinct_constants,
		      BOOL elements_in_input,
		      int max_seconds,
		      int max_megabytes,
		      BOOL flip_equations,
		      BOOL subsume,
		      FILE *fp,
		      Mace_state *msp,
		      int *rcp);

Term mace_next_model(Mace_state ms, int *rcp);

void mace_cancel(Mace_state ms);

#endif  /* conditional compilation of whole file */
