#ifndef TP_MEMORY_H
#define TP_MEMORY_H

#include "fatal.h"

/* INTRODUCTION
If you need a small chunks of memory to build data structures, this is
where to come.  We my_malloc() big blocks of memory, then break it up and
give it out as needed.
<P>
Unlike my_malloc(), the memory obtained by tp_alloc() cannot be
freed at this level of abstraction or to the operating system.
Instead, it is assumed that the caller is some
kind of get_node() routine, which, along with the corresponding
free_node() routine, manages a list of available nodes.
*/

/* Public definitions */

#define DEFAULT_MAX_MEGS            24  /* Change with set_max_megs(). */
#define DEFAULT_MALLOC_BYTES 1024*1024  /* Change with set_malloc_bytes(). */

/* End of public definitions */

/* Public function prototypes from memory.c */

void *my_malloc(int number_of_bytes);

void my_free(void *p);

void *tp_alloc(size_t n);

void *tp_calloc(size_t n);

int kbytes_malloced(void);

void set_max_megs(int megs);

void set_malloc_bytes(int megs);

#endif  /* conditional compilation of whole file */
