#include "header.h"

#include "random.h"
#include "clausify.h"
#include "pindex.h"
#include "mace.h"
#include "interp.h"
#include "demod.h"
#include "resolve.h"
#include "paramod.h"
#include "ioutil.h"
#include "hash.h"
#include "clauseid.h"
#include "subsume.h"
#include "lindex.h"
#include "clause2.h"

#define N 100
#define Length 5
#define Range 3

#define M 100

void random_term_test()
{
  int i;
  Term t;

  for (i = 0; i < 100; i++) {
    t = random_term(10, 10, 10, 10, 10, 4);
    p_term(t);
  }
}  /* random_term_test */

void btm_test()
{
  Term a[M];
  int i, j, n;
  Context c1 = get_context();
  Term t1, t2, t3;
  Btm_state bt;

  set_assoc_comm("f0");

  for (i = 0; i < M; i++) {
    a[i] = random_complex_term(2, 2, 0, 1, 0, 4);
    ac_canonical(a[i], -1);
    printf("%3d: ", i); p_term(a[i]);
  }

  for (i = 0; i < M; i++) {
    t1 = a[i];
    for (j = i+1; j < M; j++) {
      if (i == 33 && j == 49)
	printf("stop here\n");
      t2 = a[j];
      printf("\nattempting btm: "); p_term(t1);
      printf("              : "); p_term(t2);
      bt = match_bt_first(t1, c1, t2, 0);
      n = 0;
      while (bt != NULL) {
	n++;
	if (i == 8 && j == 35 && n == 3)
	  p_btm_state(bt);
	p_context(c1);
	t3 = apply(t1, c1);
	printf("apply substitution: "); p_term(t3);
	zap_term(t3);
	bt = match_bt_next(bt);
      }
      printf("                                      <%d,%d>: %d\n", i, j, n);
    }
  }
  for (i = 0; i < M; i++)
    zap_term(a[i]);

  free_context(c1);
  fprint_symbols_mem(stdout, 1);
  fprint_glist_mem(stdout, 0);
  fprint_term_mem(stdout, 0);
  fprint_unify_mem(stdout, 0);
  fprint_btm_mem(stdout, 0);
}  /* btm_test */

void btu_test()
{
  Term a[M];
  int i, j, n;
  Context c1 = get_context();
  Context c2 = get_context();
  Term t1, t2, t3, t4;
  Btu_state bt;

  set_assoc_comm("f0");

  for (i = 0; i < M; i++) {
    a[i] = random_complex_term(2, 2, 0, 1, 0, 2);
    printf("%3d: ", i); p_term(a[i]);
  }

  for (i = 0; i < M; i++) {
    t1 = a[i];
    for (j = i+1; j < M; j++) {
      if (i == 45 && j == 47)
	printf("stop here\n");
      t2 = a[j];
      printf("\nattempting btu: "); p_term(t1);
      printf("              : "); p_term(t2);
      bt = unify_bt_first(t1, c1, t2, c2);
      n = 0;
      while (bt != NULL) {
	n++;
	p_context(c1);
	p_context(c2);
	t3 = apply(t1, c1);
	t4 = apply(t2, c2);
	printf("apply substitution: "); p_term(t3);
	printf("                  : "); p_term(t4);
	zap_term(t3);
	zap_term(t4);
	bt = unify_bt_next(bt);
      }
      printf("                                      <%d,%d>: %d\n", i, j, n);
    }
  }
  for (i = 0; i < M; i++)
    zap_term(a[i]);

  free_context(c1);
  free_context(c2);
  fprint_symbols_mem(stdout, 1);
  fprint_glist_mem(stdout, 0);
  fprint_term_mem(stdout, 0);
  fprint_unify_mem(stdout, 0);
  fprint_btu_mem(stdout, 0);
}  /* btu_test */

void unify_test()
{
  Term a[M];
  int i, j;
  Context c1 = get_context();
  Context c2 = get_context();
  Trail tr;
  Term t1, t2, t3, t4;

  for (i = 0; i < M; i++) {
    a[i] = random_nonvariable_term(2, 1, 1, 0, 0, 3);
  }
  for (i = 0; i < M; i++) {
    t1 = a[i];
    for (j = i+1; j < M; j++) {
      t2 = a[j];
      printf("\nattempting unify: "); p_term(t1);
      printf("                : "); p_term(t2);
      tr = NULL;
      if (unify(t1, c1, t2, c2, &tr)) {
	p_context(c1);
	p_context(c2);
	p_trail(tr);
	t3 = apply(t1, c1);
	t4 = apply(t2, c2);
	printf("apply substitution: "); p_term(t3);
	printf("                  : "); p_term(t4);
	undo_subst(tr);
	zap_term(t3);
	zap_term(t4);
      }
      else
	printf("unify fails\n");
    }
  }
}  /* unify_test */

void fpa_state_test()
{
  int i, b[M];
  Term a[M];
  Fpa_index idx = fpa_init_index(1000);

  set_commutative("f0");

  for (i = 0; i < M; i++) {
    Term t = random_complex_term(1, 1, 0, 1, 0, 3);
    a[i] = t;
    t->INDEX_ID = 100+i;  /* cannot be 0 */
    fpa_insert(t, idx);
  }
  /* printf("\nFPA Trie:\n");  p_fpa_trie(idx->root); */

  for (i = 0; i < M; i++) {
    printf("%3d: ", i); p_term(a[i]);
  }

  for (i = 3; i < M; i++) {
    p_fpa_query(a[i], VARIANT, idx);
    p_fpa_query(a[i], GENERALIZATION, idx);
    p_fpa_query(a[i], UNIFY, idx);
    p_fpa_query(a[i], INSTANCE, idx);
    p_fpa_query(a[i], IDENTICAL, idx);
  }

  random_permutation(b, M);
  for (i = 0; i < M; i++) {
    fpa_delete(a[b[i]], idx);
  }

  for (i = 0; i < M; i++) {
    zap_term(a[i]);
  }

  zap_fpa_index(idx);

  fprint_term_mem(stdout, 1);
  fprint_fpa_mem(stdout, 0);
  fprint_symbols_mem(stdout, 0);
  fprint_glist_mem(stdout, 0);

}  /* fpa_state_test */

void mate_count(Term t, int type, Mindex idx, int *mp, int *up)
{
  Mindex_pos pos;
  Term t2;
  Context cq;
  Context cf;
  int found[M+1];
  int i;

  *mp = *up = 0;
  for (i = 0; i < M+1; i++)
    found[i] = 0;
  
  if (type == UNIFY || type == INSTANCE || type == VARIANT)
    cq = get_context();
  else
    cq = NULL;

  if (type == UNIFY || type == GENERALIZATION)
    cf = get_context();
  else
    cf = NULL;

  t2 = mindex_retrieve_first(t, idx, type, cq, cf, FALSE, &pos);
  while (t2 != NULL) {
    (*up)++;
    if (FALSE && t->INDEX_ID == 3) {
      printf("mate %d: ", t2->INDEX_ID); p_term(t2);
      p_context(cf);
    }
    if (!found[t2->INDEX_ID]) {
      found[t2->INDEX_ID] = 1;
      (*mp)++;
    }
#if 0
    else {
      printf("refound mate %d: ", t2->INDEX_ID); p_term(t2);
    }
#endif
    t2 = mindex_retrieve_next(pos);
  }
  if (cq != NULL)
    free_context(cq);
  if (cf != NULL)
    free_context(cf);

}  /* mate_count */

void compare_two(Term t, int query_type, Mindex idx1, Mindex idx2)
{
  int m1, m2, u1, u2;
  char *s;

  mate_count(t, query_type, idx1, &m1, &u1);
  mate_count(t, query_type, idx2, &m2, &u2);

  switch (query_type) {
  case UNIFY:          s = "UNIFY         "; break;
  case INSTANCE:       s = "INSTANCE      "; break;
  case GENERALIZATION: s = "GENERALIZATION"; break;
  case VARIANT:        s = "VARIANT       "; break;
  case IDENTICAL:      s = "IDENTICAL     "; break;
  default:             s = "???           "; break;
  }
  if (u1 != u2) {
    fprintf(stderr, "ERROR-u\n");
    printf("ERROR-u: ");
  }
  if (m1 != m2) {
    fprintf(stderr, "ERROR-m\n");
    printf("ERROR-m: ");
  }
#if 1
  printf("%s (%d,%d) (%d,%d)\n", s, m1, u1, m2, u2);
#endif
}  /* compare_two */

void index_test(void)
{
  int i;
  Term a[M], t;
  int b[M];
  Mindex fpa_index = mindex_init(FPA, ORDINARY_UNIF, 1);
  Mindex linear_index = mindex_init(LINEAR, ORDINARY_UNIF, 0);
  Mindex discrim_index = mindex_init(DISCRIM, ORDINARY_UNIF, 0);
  Mindex discrim_bind_index = mindex_init(DISCRIM_BIND, ORDINARY_UNIF, 0);

  for (i = 0; i < M; i++) {
    a[i] = random_nonvariable_term(3, 2, 1, 1, 0, 5);
    t = a[i]; 
    mindex_insert(fpa_index, t);
    mindex_insert(linear_index, t);
    mindex_insert(discrim_index, t);
    mindex_insert(discrim_bind_index, t);
  }

#if 0
  fprint_mindex(stdout, fpa_index);
  fprint_mindex(stdout, linear_index);
  fprint_mindex(stdout, discrim_index);
  fprint_mindex(stdout, discrim_bind_index);
#endif

  for (i = 0; i < M; i++) {
    t = a[i];
    printf("%3d: ", t->INDEX_ID); p_term(t);

    /* Compare fpa against linear for all types of mate. */

    compare_two(t, UNIFY, fpa_index, linear_index);
    compare_two(t, INSTANCE, fpa_index, linear_index);
    compare_two(t, GENERALIZATION, fpa_index, linear_index);
    compare_two(t, VARIANT, fpa_index, linear_index);
    compare_two(t, IDENTICAL, fpa_index, linear_index);

    /* Compare discrim and discrim_bind against linear for GEN */

    compare_two(t, GENERALIZATION, discrim_index, linear_index);
    compare_two(t, GENERALIZATION, discrim_bind_index, linear_index);
  }

#if 0
  index_destroy(fpa_index);
  index_destroy(linear_index);
  index_destroy(discrim_index);
  index_destroy(discrim_bind_index);
#else
  random_permutation(b, M);
  for (i = 0; i < M; i++) {
    mindex_delete(fpa_index, a[b[i]]);
    mindex_delete(linear_index, a[b[i]]);
    mindex_delete(discrim_index, a[b[i]]);
    mindex_delete(discrim_bind_index, a[b[i]]);
  }
  mindex_free(fpa_index);
  mindex_free(linear_index);
  mindex_free(discrim_index);
  mindex_free(discrim_bind_index);
#endif

  for (i = 0; i < M; i++) {
    zap_term(a[i]);
  }

  fprint_symbols_mem(stdout, 1);
  fprint_glist_mem(stdout, 0);
  fprint_term_mem(stdout, 0);
  fprint_unify_mem(stdout, 0);
  fprint_fpa_mem(stdout, 0);
  fprint_discrim_mem(stdout, 0);
  fprint_mindex_mem(stdout, 0);
}  /* index_test */

void btu_index_test(void)
{
  int i;
  Term a[M], t;
  int b[M];

  Mindex linear_bt_index = mindex_init(LINEAR, BACKTRACK_UNIF, 0);
  Mindex fpa_bt_index = mindex_init(FPA, BACKTRACK_UNIF, 100);

  set_assoc_comm("f0");

  for (i = 0; i < M; i++) {
    a[i] = random_complex_term(2, 2, 1, 1, 0, 4);
    t = a[i]; 
    mindex_insert(fpa_bt_index, t);
    mindex_insert(linear_bt_index, t);
  }

#if 0
  fprint_mindex(stdout, fpa_bt_index);
#endif

  for (i = 0; i < M; i++) {
    t = a[i];
    printf("%3d: ", t->INDEX_ID); p_term(t);

    compare_two(t, UNIFY, fpa_bt_index, linear_bt_index);
  }

#if 0
  index_destroy(fpa_bt_index);
  index_destroy(linear_bt_index);
#else
  random_permutation(b, M);
  for (i = 0; i < M; i++) {
    mindex_delete(fpa_bt_index, a[b[i]]);
    mindex_delete(linear_bt_index, a[b[i]]);
  }
  mindex_free(fpa_bt_index);
  mindex_free(linear_bt_index);
#endif

  for (i = 0; i < M; i++) {
    zap_term(a[i]);
  }

  fprint_symbols_mem(stdout, 1);
  fprint_glist_mem(stdout, 0);
  fprint_term_mem(stdout, 0);
  fprint_unify_mem(stdout, 0);
  fprint_fpa_mem(stdout, 0);
  fprint_discrim_mem(stdout, 0);
  fprint_mindex_mem(stdout, 0);
}  /* btu_index_test */

/* #define M 3 */

void btm_index_test(void)
{
  int i;
  Term a[M], t;
  int b[M];

  Mindex linear_bt_index = mindex_init(LINEAR, BACKTRACK_UNIF, 0);
  Mindex fpa_bt_index = mindex_init(FPA, BACKTRACK_UNIF, 100);
  Mindex discrim_bt_index = mindex_init(DISCRIM, BACKTRACK_UNIF, 0);
  Mindex discrimb_bt_index = mindex_init(DISCRIM_BIND, BACKTRACK_UNIF, 0);

  set_commutative("f0");

  for (i = 0; i < M; i++) {
    a[i] = random_complex_term(2, 2, 1, 1, 0, 4);
    t = a[i]; 
    mindex_insert(linear_bt_index, t);
    mindex_insert(fpa_bt_index, t);
    mindex_insert(discrim_bt_index, t);
    mindex_insert(discrimb_bt_index, t);
  }

#if 0
  fprint_mindex(stdout, fpa_bt_index);
#endif

  for (i = 0; i < M; i++) {
    t = a[i];
    printf("%3d: ", t->INDEX_ID); p_term(t);

    /* compare_two(t, GENERALIZATION, linear_bt_index, fpa_bt_index); */
    /* compare_two(t, GENERALIZATION, fpa_bt_index, discrim_bt_index); */
    compare_two(t, GENERALIZATION, discrim_bt_index, discrimb_bt_index);
    /* compare_two(t, GENERALIZATION, linear_bt_index, discrim_bt_index); */
  }

#if 0
  index_destroy(linear_bt_index);
  index_destroy(fpa_bt_index);
  index_destroy(discrim_bt_index);
  index_destroy(discrimb_bt_index);
#else
  random_permutation(b, M);
  for (i = 0; i < M; i++) {
    mindex_delete(linear_bt_index, a[b[i]]);
    mindex_delete(fpa_bt_index, a[b[i]]);
    mindex_delete(discrim_bt_index, a[b[i]]);
    mindex_delete(discrimb_bt_index, a[b[i]]);
  }
  mindex_free(linear_bt_index);
  mindex_free(fpa_bt_index);
  mindex_free(discrim_bt_index);
  mindex_free(discrimb_bt_index);
#endif

  for (i = 0; i < M; i++) {
    zap_term(a[i]);
  }

  fprint_symbols_mem(stdout, 1);
  fprint_glist_mem(stdout, 0);
  fprint_term_mem(stdout, 0);
  fprint_unify_mem(stdout, 0);
  fprint_fpa_mem(stdout, 0);
  fprint_discrim_mem(stdout, 0);
  fprint_mindex_mem(stdout, 0);
}  /* btm_index_test */

/* #define M 23 */

void discrim_test(void)
{
  int i;
  Term a[M], t;
  int b[M];

  Mindex fpa_index = mindex_init(FPA, ORDINARY_UNIF, 100);
  Mindex discrim_index = mindex_init(DISCRIM, ORDINARY_UNIF, 0);
  Mindex discrim_bind_index = mindex_init(DISCRIM_BIND, ORDINARY_UNIF, 0);

  set_commutative("f0");
  set_commutative("f1");

  for (i = 0; i < M; i++) {
    a[i] = random_complex_term(1, 1, 0, 2, 0, 5);
    t = a[i];
    mindex_insert(fpa_index, t);
    mindex_insert(discrim_index, t);
    mindex_insert(discrim_bind_index, t);
  }

  fprint_mindex(stdout, discrim_bind_index);

  for (i = 0; i < M; i++) {
    t = a[i];
    printf("%3d: ", t->INDEX_ID); p_term(t);
    /* compare_two(t, GENERALIZATION, fpa_index, discrim_index); */
    compare_two(t, GENERALIZATION, fpa_index, discrim_bind_index);
  }

  random_permutation(b, M);
  for (i = 0; i < M; i++) {
    mindex_delete(fpa_index, a[b[i]]);
    mindex_delete(discrim_index, a[b[i]]);
    mindex_delete(discrim_bind_index, a[b[i]]);
  }

  mindex_free(fpa_index);
  mindex_free(discrim_index);
  mindex_free(discrim_bind_index);

  for (i = 0; i < M; i++) {
    zap_term(a[i]);
  }

  fprint_symbols_mem(stdout, 1);
  fprint_glist_mem(stdout, 0);
  fprint_term_mem(stdout, 0);
  fprint_unify_mem(stdout, 0);
  fprint_fpa_mem(stdout, 0);
  fprint_discrim_mem(stdout, 0);
  fprint_mindex_mem(stdout, 0);
}  /* discrim_test */

void strbuf_test(void)
{
  String_buf sb = get_string_buf();
  int i;
  char s[10];

  for (i = 0; i < 26; i++) {
    sprintf(s, "%c-", 'a' + i);
    sb_append(sb, s);
  }
  p_sb(sb);
  zap_string_buf(sb);

  {
    String_buf sb = get_string_buf();
    String_buf sb2 = get_string_buf();
    Term t = random_term(5, 5, 10, 10, 10, 5);
    sprint_term(sb, t);
    p_sb(sb);
    p_term(t);

    sb_append(sb2, "abcdefghijklmnopqrstuvwxyz");
    sb_cat_copy(sb, sb2);
    p_sb(sb);

    for (i = 0; i < 100; i++) {
      Term t = random_term(5, 5, 10, 10, 10, 3);
      String_buf sb3 = get_string_buf();
      char *str;
      sprint_term(sb3, t);
      str = sb_to_new_string(sb3);
      printf("new string=%s\n", str);
      my_free(str);
      zap_string_buf(sb3);
    }

    zap_string_buf(sb);
    zap_string_buf(sb2);
    p_strbuf_mem(1);
  }
}  /* strbuf_test */

void time_test()
{
  int i, j, k;
  int c[100];
  unsigned u1, user, sys;
  int w;
  wall_init();

  for (i = 0; i < 100; i++) {
    char name[100];
    sprintf(name, "#%d", i);
    c[i] = clock_init(name);
    clock_start(c[i]);
    if (i % 20 == 0) {
      /* spin a bit */
      k = 0;
      for (j = 0; j < 1000000; j++)
	k += i+j;
    }
  }

  for (i = 0; i < 100; i++) {
    clock_stop(c[i]);
    u1 = clock_value(c[i]);
  }

  fprint_all_clocks(stdout);

  w = wall_time();
  user = user_time();
  sys = system_time();

  printf("c[10]=%d, val=%u, wall=%d, user=%u, sys=%u\n", c[10], u1, w, user, sys);
}  /* time_test */

void clause_test()
{
  int m = 20;
  int i, wt;
  Clause c, c1, c2;
  Pair_index pi = init_pair_index(100);

  Clist list = clist_init("my first list!");

  for (i = 0; i < m; i++) {
    c = random_clause(10, 10, 10, 10, 10, 2, 5);
    assign_clause_id(c);
    wt = clause_length(c);
    insert_pair_index(c, wt, pi);
    clist_append(c, list);
  }

  fprint_clist(stdout, list);

  i = 0;
  retrieve_pair(pi, &c1, &c2);
  while (c1 != NULL) {
    i++;
    printf("\npair %d\n", i);
    p_clause(c1);
    p_clause(c2);
    retrieve_pair(pi, &c1, &c2);
  }

  for (i = 1; i <= m; i++) {
    c = find_clause_by_id(i);
    wt = clause_length(c);
    delete_pair_index(c, wt, pi);
    unassign_clause_id(c);
    clist_remove(c, list);
    zap_clause(c);
  }

  clist_free(list);

  zap_pair_index(pi);

  fprint_glist_mem(stdout, 1);
  fprint_term_mem(stdout, 0);
  fprint_clause_mem(stdout, 0);
  fprint_clist_mem(stdout, 0);
  fprint_pindex_mem(stdout, 0);
}  /* clause_test */

void pindex_test()
{
  int m = 20;
  int i, wt;
  Clause c, c1, c2;
  Pair_index pi = init_pair_index(100);

  Clist list = clist_init("my first list!");

  for (i = 0; i < m; i++) {
    c = random_clause(10, 10, 10, 10, 10, 2, 5);
    assign_clause_id(c);
    wt = clause_length(c);
    insert_pair_index(c, wt, pi);
    clist_append(c, list);
  }

  fprint_clist(stdout, list);

  i = 0;
  retrieve_pair(pi, &c1, &c2);
  while (c1 != NULL) {
    i++;
    retrieve_pair(pi, &c1, &c2);
  }
  printf("Pairs retrieved, first pass: %d.\n", i);
  printf("exhausted: %s.\n", pairs_exhausted(pi) ? "TRUE" : "FALSE");

  c = random_clause(10, 10, 10, 10, 10, 2, 5);
  assign_clause_id(c);
  wt = clause_length(c);
  insert_pair_index(c, wt, pi);
  printf("exhausted: %s.\n", pairs_exhausted(pi) ? "TRUE" : "FALSE");

  i = 0;
  retrieve_pair(pi, &c1, &c2);
  while (c1 != NULL) {
    i++;
    retrieve_pair(pi, &c1, &c2);
  }
  printf("Pairs retrieved, second pass: %d.\n", i);

  for (i = 1; i <= m; i++) {
    c = find_clause_by_id(i);
    wt = clause_length(c);
    delete_pair_index(c, wt, pi);
    unassign_clause_id(c);
    clist_remove(c, list);
    zap_clause(c);
  }

  clist_free(list);

  zap_pair_index(pi);

  fprint_glist_mem(stdout, 1);
  fprint_term_mem(stdout, 0);
  fprint_clause_mem(stdout, 0);
  fprint_clist_mem(stdout, 0);
  fprint_pindex_mem(stdout, 0);
}  /* pindex_test */

void term_flag_test(void)
{
  int i, j;
  Term a[10], t;

  for (i = -2; i < 10; i++) {
    t = random_term(5, 5, 10, 10, 10, 3);
    a[i] = t;
    term_flag_set(t, i);
    printf("\nflags=%d ", t->private_flags); p_term(t);

    for (j = -2; j < 10; j++)
      printf("flag %d: %d\n", j, term_flag(t, j));

    term_flag_clear(t, i);
    printf("flags=%d ", t->private_flags); p_term(t);
  }
  
}  /* term_flag_test */

void declare_ops(void) {
  set_parse_type(",",       999, INFIX_RIGHT);

  set_parse_type(IFF_SYM,   800, INFIX);
  set_parse_type(IMP_SYM,   800, INFIX);
  set_parse_type(IMPBY_SYM, 800, INFIX);
  set_parse_type(OR_SYM,    790, INFIX_RIGHT);
  set_parse_type(AND_SYM,   780, INFIX_RIGHT);
  set_parse_type(NOT_SYM,   770, PREFIX);

  set_parse_type("=",   700, INFIX);
  set_parse_type("!=",  700, INFIX);

  set_parse_type("+",   500, INFIX_LEFT);
  set_parse_type("*",   400, INFIX_LEFT);
  set_parse_type("\'",  300, POSTFIX);
}  /* declare_ops */

void read_term_test() {
  Term t;

  declare_ops();

  t = read_term(stdin, stdout);
  while (t != NULL) {
    fwrite_term_nl(stdout, t);
    fprint_term(stdout, t); fprintf(stdout, ".\n\n");
    zap_term(t);
    t = read_term(stdin, stdout);
  }

  fprint_strbuf_mem(stdout, 1);
  fprint_term_mem(stdout, 0);
  fprint_parse_mem(stdout, 0);
}  /* read_term_test */

/* #define M 200 */

void random_op_test(void)
{
  Term a[M], b[M];
  int i;
  FILE *fp;

  declare_ops();

  fp = fopen("wacky-terms", "w");
  for (i = 0; i < M; i++) {
    a[i] = random_op_term(8);
    fwrite_term_nl(fp, a[i]);
    /* fprint_term(fp, a[i]); fprintf(fp, ".\n\n"); */
  }
  fclose(fp);

  fp = fopen("wacky-terms", "r");

  for (i = 0; i < M; i++) {
    b[i] = read_term(fp, stdout);
    if (b[i] != NULL && term_ident(a[i], b[i])) {
      printf("%d ok\n", i);
    }
    else {
      printf("\n==============%d bad news========\n\n", i);
      if (b[i] != NULL) {
	fwrite_term_nl(stdout, a[i]);
	fwrite_term_nl(stdout, b[i]);
	printf("\n-------\n");
	fprint_term(stdout, a[i]); fprintf(stdout, ".\n");
	fprint_term(stdout, b[i]); fprintf(stdout, ".\n");
      }
    }
    fflush(stdout);
  }
  fclose(fp);
  for (i = 0; i < M; i++) {
    zap_term(a[i]);
    if (b[i] != NULL)
      zap_term(b[i]);
  }

  fprint_strbuf_mem(stdout, 1);
  fprint_term_mem(stdout, 0);
  fprint_parse_mem(stdout, 0);
}  /* random_op_test */

void clause_trans_test(void)
{
  Term t, t2;
  Clause c;
  int ok;

  declare_ops();

  t = read_term(stdin, stdout);
  while (t != NULL) {
    printf("fwrite_term_nl: "); fwrite_term_nl(stdout, t);
    printf("fprint_term: "); fprint_term(stdout, t); fprintf(stdout, ".\n");
    c = term_to_clause(t);
    printf("fprint_clasue:        "); fprint_clause(stdout, c);
    ok = clause_set_variables(c, 10);
    if (ok) {
      printf("fprint_clasue: (vars) "); fprint_clause(stdout, c);
    }
    else
      printf("too many variables.\n");

    t2 = clause_to_term(c);
    if (!term_ident(t, t2)) {
      printf("bad term: "); fprint_term(stdout, t2); fprintf(stdout, ".\n");
    }
    zap_term(t);
    zap_term(t2);
    zap_clause(c);

    t = read_term(stdin, stdout);
  }

  fprint_strbuf_mem(stdout, 1);
  fprint_term_mem(stdout, 0);
  fprint_parse_mem(stdout, 0);
  fprint_clause_mem(stdout, 0);
}  /* clause_trans_test */

void formula_test(void)
{
  Term t;
  Formula f;

  declare_ops();

  t = read_term(stdin, stdout);
  while (t != NULL) {
    printf("fwrite_term_nl: "); fwrite_term_nl(stdout, t);

    f = term_to_formula(t);
    printf("formula:      "); fwrite_formula(stdout, f);

    f = clausify_prepare(f);
    printf("clause_prep:  "); fwrite_formula(stdout, f);

#if 0
    f = unique_quantified_vars(f);
    printf("unique qvars: "); fwrite_formula(stdout, f);

    f = remove_universal_quantifiers(f);
    printf("remove uvars: "); fwrite_formula(stdout, f);

    f = formula_simplify(f);
    printf("simplified:  "); fwrite_formula(stdout, f);

    f = nnf(f, CONJUNCTION);
    printf("nnf formula: "); fwrite_formula(stdout, f);

    f = skolemize(f);
    printf("skolemize:   "); fwrite_formula(stdout, f);

    f = cnf(f, 1);
    printf("cnf formula: "); fwrite_formula(stdout, f);

    f = dnf(f, 1);
    printf("dnf formula: "); fwrite_formula(stdout, f);
#endif

    zap_term(t);
    zap_formula(f);

    t = read_term(stdin, stdout);
  }

  fprint_strbuf_mem(stdout, 1);
  fprint_term_mem(stdout, 0);
  fprint_parse_mem(stdout, 0);
  fprint_formula_mem(stdout, 0);
}  /* formula_test */

void clausify_test(void)
{
  Term t;
  Formula f;
  Clist lst;
  Clist_pos p;

  declare_ops();

  t = read_term(stdin, stdout);
  while (t != NULL) {
    printf("fwrite_term_nl: "); fwrite_term_nl(stdout, t);

    f = term_to_formula(t);
    printf("formula:      "); fwrite_formula(stdout, f);

    lst = clausify(f);
    printf("clauses:\n");
    for (p = lst->first; p != NULL; p = p->next)
      p_clause(p->c);

    clist_zap(lst);
    zap_term(t);

    t = read_term(stdin, stdout);
  }

  fprint_strbuf_mem(stdout, 1);
  fprint_parse_mem(stdout, 0);
  fprint_term_mem(stdout, 0);
  fprint_formula_mem(stdout, 0);
  fprint_clause_mem(stdout, 0);
  fprint_clist_mem(stdout, 0);
}  /* clausify_test */

void interp_test(void)
{
  Term t;
  Interp p;
  Clause c;
  int rc1, ok, ck1;

  declare_ops();

  t = read_term(stdin, stdout);
  fwrite_term_nl(stdout, t);

  p = compile_interp(t);
  p_interp(p);
  zap_term(t);

  t = read_term(stdin, stdout);
  c = term_to_clause(t);
  ok = clause_set_variables(c, 20);
  zap_term(t);
  p_clause(c);

  ck1 = clock_init("old");
  clock_start(ck1);
  rc1 = eval_clause(c, p);
  clock_stop(ck1);
  printf("Clause is %s.\n", rc1 ? "TRUE" : "FALSE");

  fprint_all_clocks(stdout);

  zap_clause(c);

  zap_interp(p);

  fprint_strbuf_mem(stdout, 1);
  fprint_parse_mem(stdout, 0);
  fprint_term_mem(stdout, 0);
  fprint_formula_mem(stdout, 0);
  fprint_clause_mem(stdout, 0);
  fprint_clist_mem(stdout, 0);
  fprint_interp_mem(stdout, 0);
}  /* interp_test */

void interp_formula_test(void)
{
  Term t;
  Interp p;
  Formula f;
  int rc1, ck1;

  declare_ops();

  t = read_term(stdin, stdout);
  fwrite_term_nl(stdout, t);

  p = compile_interp(t);
  p_interp(p);
  zap_term(t);

  t = read_term(stdin, stdout);

  f = term_to_formula(t);
  f = eliminate_rebinding(f);

  zap_term(t);
  fwrite_formula(stdout, f); 

  ck1 = clock_init("");
  clock_start(ck1);
  rc1 = eval_formula(f, p);
  clock_stop(ck1);
  printf("Formula is %s.\n", rc1 ? "TRUE" : "FALSE");

  fprint_all_clocks(stdout);

  zap_formula(f);

  zap_interp(p);

  fprint_strbuf_mem(stdout, 1);
  fprint_parse_mem(stdout, 0);
  fprint_term_mem(stdout, 0);
  fprint_formula_mem(stdout, 0);
  fprint_clause_mem(stdout, 0);
  fprint_clist_mem(stdout, 0);
  fprint_interp_mem(stdout, 0);
}  /* interp_formula_test */

void demod_test(void)
{
  Glist terms, g;
  Clist demodulators;
  Clist_pos p;
  Mindex idx;

  declare_standard_parse_types();
  set_assoc_comm("*");
  set_assoc_comm("+");

  /* Read list of demodulators. */

  demodulators = read_clause_list(stdin, stdout, "demodulators", TRUE);
  for (p = demodulators->first; p != NULL; p = p->next) {
    Clause c = p->c;
    Literal lit;
    for (lit = c->first_literal; lit != NULL; lit = lit->next)
      ac_canonical(lit->atom, -1);
  }
  fwrite_clause_list(stdout, demodulators, TRUE);

  /* Read list of terms to be demodulated. */

  terms = read_term_list(stdin, stdout);
  fwrite_term_list(stdout, terms, "terms");

  /* Index the demodulators. */

  idx = mindex_init(DISCRIM, BACKTRACK_UNIF, 0);
  for (p = demodulators->first; p != NULL; p = p->next) {
    /* assume pos equality unit */
    Term alpha;
    alpha = p->c->first_literal->atom->args[0];
    mindex_insert(idx, alpha);
  }

  /* Demoulate the terms. */

  for (g = terms; g != NULL; g = g->next) {
    Term t = copy_term(g->u.v);
    Glist just = NULL;
    fprintf(stdout, "\nBefore:   "); fwrite_term_nl(stdout, t);
    ac_canonical(t, -1);
    fprintf(stdout, "AC-canon: "); fwrite_term_nl(stdout, t);
    t = demodulate(t, idx, &just);
    just = reverse_glist(just);
    fprintf(stdout,   "after:    ");  fprint_ilist(stdout, just);
    fwrite_term_nl(stdout, t);
    zap_glist(just);
    zap_term(t);
  }

  /* Unindex the demodulators. */

  for (p = demodulators->first; p != NULL; p = p->next) {
    /* assume pos equality unit */
    Term alpha = p->c->first_literal->atom->args[0];
    mindex_delete(idx, alpha);
  }
  mindex_free(idx);

  /* Free the lists. */

  clist_zap(demodulators);
  g = terms;
  while (g != NULL) {
    Glist g2 = g;
    g = g->next;
    zap_term(g2->u.v);
    free_glist(g2);
  }

  printf("Demod_attempts=%d, Demod_rewrites=%d.\n",
	 demod_attempts(), demod_rewrites());

  /* Look for memory leaks. */

  fprint_strbuf_mem(stdout, 1);
  fprint_parse_mem(stdout, 0);
  fprint_term_mem(stdout, 0);
  fprint_glist_mem(stdout, 0);
  fprint_clause_mem(stdout, 0);
  fprint_clist_mem(stdout, 0);
  fprint_discrim_mem(stdout, 0);
  fprint_fpa_mem(stdout, 0);
  fprint_unify_mem(stdout, 0);
  fprint_btu_mem(stdout, 0);
  fprint_btm_mem(stdout, 0);
  fprint_mindex_mem(stdout, 0);
}  /* demod_test */

void resolve_test(void)
{
  Clist clauses;
  Clist_pos p;
  Lindex ldx;

  declare_standard_parse_types();

  /* Read list of clauses. */

  clauses = read_clause_list(stdin, stdout, "clauses", TRUE);
  fwrite_clause_list(stdout, clauses, TRUE);

  /* Index the clauses. */

  ldx = lindex_init(FPA, ORDINARY_UNIF, 100, FPA, ORDINARY_UNIF, 100);

  for (p = clauses->first; p != NULL; p = p->next)
    lindex_insert(ldx, p->c);

  /* Make inferences. */

  for (p = clauses->first; p != NULL; p = p->next) {
    Glist resolvents;
    printf("\ngiven: "); fwrite_clause(stdout, p->c, FALSE);
    /* binary_factors(p->c); */
    resolvents = ur_resolution(p->c, ldx);
    while (resolvents != NULL) {
      Clause c = resolvents->u.v;
      renumber_variables(c, MAX_VARS);
      assign_clause_id(c);
      fwrite_clause(stdout, c, TRUE);
      zap_just(c->justification);
      unassign_clause_id(c);
      zap_clause(c);
      resolvents = glist_pop(resolvents);
    }
  }

  /* Unindex the clauses and remove from clause ID table. */

  for (p = clauses->first; p != NULL; p = p->next) {
    lindex_delete(ldx, p->c);
    unassign_clause_id(p->c);
  }

  lindex_destroy(ldx);

  /* Free the lists. */

  clist_zap(clauses);

  /* Look for memory leaks. */

  fprint_strbuf_mem(stdout, 1);
  fprint_parse_mem(stdout, 0);
  fprint_term_mem(stdout, 0);
  fprint_glist_mem(stdout, 0);
  fprint_clause_mem(stdout, 0);
  fprint_clist_mem(stdout, 0);
  fprint_discrim_mem(stdout, 0);
  fprint_fpa_mem(stdout, 0);
  fprint_unify_mem(stdout, 0);
  fprint_btu_mem(stdout, 0);
  fprint_btm_mem(stdout, 0);
  fprint_mindex_mem(stdout, 0);
  fprint_lindex_mem(stdout, 0);
  fprint_clash_mem(stdout, 0);
}  /* resolve_test */

void paramod_test(void)
{
  Clist clauses;
  Clist_pos p;
  Clause c, d;

  declare_standard_parse_types();
  set_assoc_comm("+");
  set_assoc_comm("*");

  /* Read list of clauses. */

  clauses = read_clause_list(stdin, stdout, "clauses", TRUE);
  fwrite_clause_list(stdout, clauses, TRUE);

  /* Make inferences. */


  c = clauses->first->c;
  d = clauses->first->next->c;

  {
    Glist p1 = para_from_into(c, d, PARA_ALL,            FALSE, FALSE);
    Glist p2 = para_from_into(c, d, PARA_ALL,            TRUE,  FALSE);
    Glist p3 = para_from_into(c, d, PARA_TOP_ONLY,       FALSE, TRUE);
    Glist p4 = para_from_into(c, d, PARA_TOP_ONLY,       TRUE,  TRUE);
    Glist p5 = para_from_into(d, c, PARA_ALL_EXCEPT_TOP, FALSE, FALSE);
    Glist p6 = para_from_into(d, c, PARA_ALL_EXCEPT_TOP, TRUE,  FALSE);
  }

  /* Free the list. */

  for (p = clauses->first; p != NULL; p = p->next)
    unassign_clause_id(p->c);

  clist_zap(clauses);

  /* Look for memory leaks. */

  fprint_strbuf_mem(stdout, 1);
  fprint_parse_mem(stdout, 0);
  fprint_term_mem(stdout, 0);
  fprint_glist_mem(stdout, 0);
  fprint_clause_mem(stdout, 0);
  fprint_clist_mem(stdout, 0);
  fprint_discrim_mem(stdout, 0);
  fprint_fpa_mem(stdout, 0);
  fprint_unify_mem(stdout, 0);
  fprint_btu_mem(stdout, 0);
  fprint_btm_mem(stdout, 0);
  fprint_mindex_mem(stdout, 0);
  fprint_clash_mem(stdout, 0);
  fprint_hash_mem(stdout, 0);
}  /* paramod_test */

void lrpo_test(void)
{
  Glist terms, p;

  declare_standard_parse_types();
  terms = read_term_list(stdin, stdout);

  set_lex_val(str_to_sn("0", 0), 0);
  set_lex_val(str_to_sn("1", 0), 1);
  set_lex_val(str_to_sn("2", 0), 2);
  set_lex_val(str_to_sn("3", 0), 3);
  set_lex_val(str_to_sn("4", 0), 4);

  for (p = terms; p != NULL; p = p->next) {
    Term t = p->u.v;
    Term a1 = t->args[0];
    Term a2 = t->args[1];
    int ok = lrpo_multiset(a1, a2);
    fprint_term(stdout, t);
    printf("   %s\n", ok ? "TRUE" : "FALSE");
  }
}  /* lrpo_test */

void subsume_test(void)
{
  Clist clauses;
  Clist_pos p;
  Clause c, d;
  BOOL rc;

  declare_standard_parse_types();
  set_assoc_comm("+");
  set_assoc_comm("*");
  set_commutative("=");

  /* Read list of clauses. */

  clauses = read_clause_list(stdin, stdout, "clauses", TRUE);
  fwrite_clause_list(stdout, clauses, TRUE);

  /* Make inferences. */

  d = clauses->first->c;
  c = clauses->first->next->c;

  rc = subsumes_bt(c, d);
  printf("c subsumes d: %s\n", rc ? "TRUE" : "FALSE");

  for (p = clauses->first; p != NULL; p = p->next)
    unassign_clause_id(p->c);

  clist_zap(clauses);

  /* Look for memory leaks. */

  fprint_strbuf_mem(stdout, 1);
  fprint_parse_mem(stdout, 0);
  fprint_term_mem(stdout, 0);
  fprint_glist_mem(stdout, 0);
  fprint_clause_mem(stdout, 0);
  fprint_clist_mem(stdout, 0);
  fprint_discrim_mem(stdout, 0);
  fprint_fpa_mem(stdout, 0);
  fprint_unify_mem(stdout, 0);
  fprint_btu_mem(stdout, 0);
  fprint_btm_mem(stdout, 0);
  fprint_mindex_mem(stdout, 0);
  fprint_clash_mem(stdout, 0);
  fprint_hash_mem(stdout, 0);
}  /* subsume_test */

int main(int argc, char **argv)
{
  /* strbuf_test(); */
  /* fpa_state_test(); */
  /* unify_test(); */
  /* index_test(); */
  /* discrim_test(); */
   btu_test();
  /* btm_test(); */
  /* btu_index_test(); */
  /* btm_index_test(); */
  /* time_test(); */
  /* clause_test(); */
  /* pindex_test(); */
  /* term_flag_test(); */
  /* read_term_test(); */
  /* random_op_test(); */
  /* clause_trans_test(); */
  /* formula_test(); */
  /* clausify_test(); */
  /* interp_test(); */
  /* interp_formula_test(); */
  /* demod_test(); */
  /* resolve_test(); */
  /* paramod_test(); */
  /* lrpo_test(); */
  /* subsume_test(); */

  exit(0);
}  /* main */

