#include "order.h"

/*************
 *
 *    merge_sort
 *
 *************/

/* DOCUMENTATION
This is a general-purpose sorting routine.  You give it an
array of pointers to sort, an array (at least as big) for working storage,
a range of objects to sort, and a comparison function.
<P>
Here is an example of how to use it.
<PRE>
  {
    Term args[MAX_ACM_ARGS], work[MAX_ACM_ARGS];
    int n;

    < put the n terms you wish to sort into args[] >

    merge_sort((void **) args, (void **) work, 0, n-1,
               (Ordertype (*)(void*,void*)) term_compare_ncv);

    < args[] is now ordered by term_compare_ncv() >
  }
</PRE>
*/

/* PUBLIC */
void merge_sort(void *a[],    /* array to sort */
		void *w[],    /* work array */
		int start,    /* index of first element */
		int end,      /* index of last element */
		Ordertype (*comp_proc) (void *, void *))
{
  int mid, i, i1, i2, e1, e2;

  if (start < end) {
    mid = (start+end)/2;
    merge_sort(a, w, start, mid, comp_proc);
    merge_sort(a, w, mid+1, end, comp_proc);
    i1 = start; e1 = mid;
    i2 = mid+1; e2 = end;
    i = start;
    while (i1 <= e1 && i2 <= e2) {
      if ((*comp_proc)(a[i1], a[i2]) == GREATER_THAN)
	w[i++] = a[i2++];
      else
	w[i++] = a[i1++];
    }

    if (i2 > e2)
      while (i1 <= e1)
	w[i++] = a[i1++];
    else
      while (i2 <= e2)
	w[i++] = a[i2++];

    for (i = start; i <= end; i++)
      a[i] = w[i];
  }
}  /* merge_sort */

