#include "resolve.h"

/* Private definitions and types */

/*************
 *
 *   hyper_satellite_test()
 *
 *************/

static
BOOL hyper_satellite_test(Literal lit)
{
  /* Assume that lit occurs in a positive clause.
   * Check if lit is OK for resolution.
   */
  return TRUE;
}  /* hyper_satellite_test */

/*************
 *
 *   hyper_satellite()
 *
 *************/

static
Glist hyper_satellite(Clause c, Lindex idx)
{
  /* Assume c is positive. */
  Context sat_subst = get_context();
  Context nuc_subst = get_context();
  Literal slit;
  Glist resolvents = NULL;
  for (slit = c->first_literal; slit != NULL; slit = slit->next) {
    if (hyper_satellite_test(slit)) {
      /* for each acceptable nucleus */
      Term fnd_atom;
      Mindex_pos mate_pos;
      fnd_atom = mindex_retrieve_first(slit->atom, idx->neg, UNIFY,
				       sat_subst, nuc_subst,
				       FALSE, &mate_pos);
      while (fnd_atom != NULL) {
	Clause nuc = fnd_atom->container;
	Literal nuc_lit = atom_to_literal(fnd_atom);
	if (!positive_clause(nuc)) {
	  /* We have a nucleus.  Set up the clash list, then clash. */
	  Clash p = NULL;
	  Clash first = NULL;
	  Literal nlit;
	  for (nlit = nuc->first_literal; nlit != NULL; nlit = nlit->next) {
	    p = append_clash(p);
	    if (first == NULL)
	      first = p;
	    p->nuc_lit = nlit;
	    p->nuc_subst = nuc_subst;
	    if (nlit == nuc_lit) {
	      p->sat_lit = slit;
	      p->sat_subst = sat_subst;
	      p->clashable = TRUE;
	      p->clashed = TRUE;
	    }
	    else {
	      p->clashable = !nlit->sign;
	      if (p->clashable) {
		p->mate_index = idx->pos;
		p->sat_subst = get_context();
	      }
	    }
	  }  /* for each literal of nucleus */
	  clash(first, positive_clause, HYPER_RES_JUST, &resolvents);
	  zap_clash(first);
	}  /* if we have a nuc */
	fnd_atom = mindex_retrieve_next(mate_pos);
      }  /* for each mate */
    }  /* if sat is ok */
  }  /* for each literal of satellite */
  free_context(sat_subst);
  free_context(nuc_subst);
  return resolvents;
}  /* hyper_satellite */

/*************
 *
 *   hyper_nucleus()
 *
 *************/

static
Glist hyper_nucleus(Clause c, Lindex idx)
{
  /* Assume c is not positive. */
  Clash p = NULL;
  Clash first = NULL;
  Literal lit;
  Context nuc_subst = get_context();
  Glist resolvents = NULL;
  for (lit = c->first_literal; lit != NULL; lit = lit->next) {
    p = append_clash(p);
    if (first == NULL)
      first = p;
    p->clashable = !lit->sign;
    p->nuc_lit = lit;
    p->nuc_subst = nuc_subst;
    if (p->clashable) {
      p->mate_index = idx->pos;
      p->sat_subst = get_context();
    }
  }
  clash(first, positive_clause, HYPER_RES_JUST, &resolvents);
  free_context(nuc_subst);
  zap_clash(first);  /* This also frees satellite contexts. */
  return resolvents;
}  /* hyper_nucleus */

/*************
 *
 *   hyper_resolution()
 *
 *************/

/* DOCUMENTATION
Hyperresolution.
*/

/* PUBLIC */
Glist hyper_resolution(Clause c, Lindex idx)
{
  if (positive_clause(c))
    return hyper_satellite(c, idx);
  else
    return hyper_nucleus(c, idx);
}  /* hyper_resolution */

/*************
 *
 *   ur_satellite()
 *
 *************/

static
Glist ur_satellite(Clause c, Lindex idx)
{
  /* Assume C is a unit. */
  Context sat_subst = get_context();
  Context nuc_subst = get_context();
  Term fnd_atom;
  Mindex_pos mate_pos;
  Literal slit = c->first_literal;
  Glist resolvents = NULL;

  fnd_atom = mindex_retrieve_first(slit->atom, (slit->sign?idx->neg:idx->pos),
				   UNIFY, sat_subst, nuc_subst,
				   FALSE, &mate_pos);
  while (fnd_atom != NULL) {
    Clause nuc = fnd_atom->container;
    Literal fnd_lit = atom_to_literal(fnd_atom);
    if (!unit_clause(nuc)) {
      Literal target;
      for (target=nuc->first_literal; target!=NULL; target=target->next) {
	if (target != fnd_lit) {
	  Clash p = NULL;
	  Clash first = NULL;
	  Literal nlit;
	  for (nlit = nuc->first_literal; nlit != NULL; nlit = nlit->next) {
	    p = append_clash(p);
	    if (first == NULL)
	      first = p;
	    p->nuc_lit = nlit;
	    p->nuc_subst = nuc_subst;
	    if (nlit == fnd_lit) {
	      p->sat_lit = slit;
	      p->sat_subst = sat_subst;
	      p->clashable = TRUE;
	      p->clashed = TRUE;
	    }
	    else {
	      p->clashable = (nlit != target);
	      if (p->clashable) {
		p->mate_index = (nlit->sign ? idx->neg : idx->pos);
		p->sat_subst = get_context();
	      }
	    }
	  }  /* for each literal of nucleus */
	  clash(first, unit_clause, UR_RES_JUST, &resolvents);
	  zap_clash(first);
	}  
      }  /* for each target */
    }  /* if we have a nuc */
    fnd_atom = mindex_retrieve_next(mate_pos);
  }  /* for each mate */
  free_context(sat_subst);
  free_context(nuc_subst);
  return resolvents;
}  /* ur_satellite */

/*************
 *
 *   ur_nucleus()
 *
 *************/

static
Glist ur_nucleus(Clause c, Lindex idx)
{
  Literal target;
  Glist resolvents = NULL;
  for (target = c->first_literal; target != NULL; target = target->next) {
    Clash p = NULL;
    Clash first = NULL;
    Literal lit;
    Context nuc_subst = get_context();
    for (lit = c->first_literal; lit != NULL; lit = lit->next) {
      p = append_clash(p);
      if (first == NULL)
	first = p;
      p->clashable = (lit != target);
      p->nuc_lit = lit;
      p->nuc_subst = nuc_subst;
      if (p->clashable) {
	p->mate_index = (lit->sign ? idx->neg : idx->pos);
	p->sat_subst = get_context();
      }
    }
    clash(first, unit_clause, UR_RES_JUST, &resolvents);
    free_context(nuc_subst);
    zap_clash(first);  /* This also frees satellite contexts. */
  }
  return resolvents;
}  /* ur_nucleus */

/*************
 *
 *   ur_resolution()
 *
 *************/

/* DOCUMENTATION
Unit-resulting resolution.
*/

/* PUBLIC */
Glist ur_resolution(Clause c, Lindex idx)
{
  if (unit_clause(c))
    return ur_satellite(c, idx);
  else
    return ur_nucleus(c, idx);
}  /* ur_resolution */

/*************
 *
 *   binary_resolution()
 *
 *************/

/* DOCUMENTATION
Binary resolution.
*/

/* PUBLIC */
Glist binary_resolution(Clause c, Lindex idx)
{
  Literal lit;
  Glist resolvents = NULL;
  for (lit = c->first_literal; lit != NULL; lit = lit->next) {
    Clash p = NULL;
    Clash first = NULL;
    Literal l2;
    Context nuc_subst = get_context();
    for (l2 = c->first_literal; l2 != NULL; l2 = l2->next) {
      p = append_clash(p);
      if (first == NULL)
	first = p;
      p->clashable = (l2 == lit);
      p->nuc_lit = l2;
      p->nuc_subst = nuc_subst;
      if (p->clashable) {
	p->mate_index = (lit->sign ? idx->neg : idx->pos);
	p->sat_subst = get_context();
      }
    }
    clash(first, any_clause, BINARY_RES_JUST, &resolvents);
    free_context(nuc_subst);
    zap_clash(first);  /* This also frees satellite contexts. */
  }
  return resolvents;
}  /* binary_resolution */

/*************
 *
 *   binary_factors()
 *
 *************/

/* DOCUMENTATION
*/

/* PUBLIC */
Glist binary_factors(Clause c, Uniftype utype)
{
  Literal l1;
  int i = 1;
  Glist factors = NULL;
  Context subst = get_context();
  for (l1 = c->first_literal; l1 != NULL; l1 = l1->next, i++) {
    Literal l2;
    int j = i+1;
    for (l2 = l1->next; l2 != NULL; l2 = l2->next, j++) {

      if (utype == ORDINARY_UNIF) {
	Trail tr = NULL;
	if (l1->sign == l2->sign && unify(l1->atom,subst,l2->atom,subst,&tr)) {
	  Clause f = get_clause();
	  Literal l3;
	  f->justification = factor_just(c, j);
	  for (l3 = c->first_literal; l3 != NULL; l3 = l3->next) {
	    if (l3 != l2)
	      append_literal(f, apply_lit(l3, subst));
	  }
	  undo_subst(tr);
	  upward_clause_links(f);
	  factors = plist_prepend(factors, f);
	}
      }  /* ordinary unification */
      else {
	/* backtrack unification */
	Btu_state bt = unify_bt_first(l1->atom, subst, l2->atom, subst);
	while (bt != NULL) {
	  Clause f = get_clause();
	  Literal l3;
	  f->justification = factor_just(c, j);
	  for (l3 = c->first_literal; l3 != NULL; l3 = l3->next) {
	    if (l3 != l2)
	      append_literal(f, apply_lit(l3, subst));
	  }
	  upward_clause_links(f);
	  factors = plist_prepend(factors, f);
	  bt = unify_bt_next(bt);
	}
      }  /* backtrack unification */
    }
  }
  free_context(subst);
  return factors;
}  /* binary_factors */

